import json
import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from tqdm import tqdm

from lhotse import CutSet, MonoCut, fix_manifests
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.supervision import AlignmentItem, SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike


def download_notsofar1(
    target_dir: Pathlike = ".",
    parts: Tuple[str] = ("train", "dev", "test"),
    mic: str = "sdm",
    train_version: str = "240825.1_train",
    dev_version: str = "240825.1_dev1",
    test_version: str = "240629.1_eval_small_with_GT",
    force_download: Optional[bool] = False,
) -> Path:
    try:
        from huggingface_hub import snapshot_download
    except ImportError as import_error:
        raise RuntimeError(
            "huggingface_hub is required for NOTSOFAR downloads. Install it via:\n"
            "  pip install huggingface_hub\n"
        ) from import_error

    hugging_face_token = os.getenv("HF_TOKEN")
    if not hugging_face_token:
        raise RuntimeError(
            "HuggingFace token not found. Please set the HF_TOKEN environment variable. "
            "If you have set it, please restart the session. "
        )

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    for part in parts:
        if part == "train":
            subset_name = "train_set"
            version = train_version
        elif part == "dev":
            subset_name = "dev_set"
            version = dev_version
        elif part == "test":
            subset_name = "eval_set"
            version = test_version
        else:
            raise ValueError(
                f"Unknown part: {part}. Expected one of: 'train', 'dev', 'test'."
            )

        download_patterns = [
            f"benchmark-datasets/{subset_name}/{version}/MTG/*/*.json",
        ]
        if mic == "sdm":
            download_patterns.append(
                f"benchmark-datasets/{subset_name}/{version}/MTG/*/sc_*"
            )
        elif mic == "mdm":
            download_patterns.append(
                f"benchmark-datasets/{subset_name}/{version}/MTG/*/mc_*"
            )

        snapshot_download(
            repo_id="microsoft/NOTSOFAR",
            repo_type="dataset",
            local_dir=target_dir,
            force_download=bool(force_download),
            allow_patterns=download_patterns,
        )

    return target_dir


def prepare_notsofar1(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    corpus_dir = Path(corpus_dir) / "benchmark-datasets"
    if output_dir is None:
        raise ValueError("output_dir must be provided")
    output_dir = Path(output_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    os.makedirs(output_dir, exist_ok=True)

    manifests = defaultdict(dict)

    for part in _listdir_safe(corpus_dir):
        part_dir = corpus_dir / part
        manifests[part] = defaultdict(dict)

        for version in _listdir_safe(part_dir):
            version_dir = part_dir / version / "MTG"
            sc_cuts, mc_cuts = process_data(
                version_dir, word_level=False, create_word_alignment=True
            )
            manifests[part][version] = defaultdict(dict)

            if sc_cuts:
                sc_recs, sc_sups = fix_manifests(
                    *CutSet.from_cuts(sc_cuts).decompose()[:2]
                )
                tag = f"notsofar1_sdm_{part}_{version}"
                sc_recs.to_file(output_dir / f"{tag}_recordings.jsonl.gz")
                sc_sups.to_file(output_dir / f"{tag}_supervisions.jsonl.gz")
                manifests[part][version]["single_channel"] = {
                    "recordings": sc_recs,
                    "supervisions": sc_sups,
                }

            if mc_cuts:
                mc_recs, mc_sups = fix_manifests(
                    *CutSet.from_cuts(mc_cuts).decompose()[:2]
                )
                tag = f"notsofar1_mdm_{part}_{version}"
                mc_recs.to_file(output_dir / f"{tag}_recordings.jsonl.gz")
                mc_sups.to_file(output_dir / f"{tag}_supervisions.jsonl.gz")
                manifests[part][version]["multi_channel"] = {
                    "recordings": mc_recs,
                    "supervisions": mc_sups,
                }

    return manifests


def _listdir_safe(path: Pathlike) -> List[str]:
    return list(filter(lambda name: ".DS_Store" not in name, os.listdir(path)))


def process_data(
    dataset_path, word_level=False, create_word_alignment=True
) -> Tuple[List[MonoCut], List[MonoCut]]:
    meetings = sorted(_listdir_safe(dataset_path))
    sc_cuts = []
    mc_cuts = []

    for meeting in tqdm(meetings):
        meeting_root = dataset_path / meeting
        transcription_path = meeting_root / "gt_transcription.json"
        devices = sorted(
            list(
                filter(
                    lambda x: x != "close_talk" and os.path.isdir(meeting_root / x),
                    _listdir_safe(meeting_root),
                )
            )
        )

        with open(transcription_path, "r") as f:
            transcription_json = json.load(f)

        for device in devices:
            device_path = meeting_root / device
            device_id = f"{meeting}_{device}"
            is_multi_channel = "mc" in device
            if is_multi_channel:
                # We assume the channel numbers range from 0 to num_channels - 1.
                num_channels = len(_listdir_safe(device_path))
                recording = Recording.from_file(device_path / f"ch0.wav")
                recording.id = device_id
                recording.channel_ids = list(range(num_channels))
                recording.sources = [
                    AudioSource(
                        type="file",
                        channels=[i],
                        source=str(device_path / f"ch{i}.wav"),
                    )
                    for i in range(num_channels)
                ]
            else:
                recording_path = device_path / "ch0.wav"
                recording = Recording.from_file(recording_path)
                recording.id = device_id

            supervisions = []
            for segment in transcription_json:
                speaker_id = segment["speaker_id"]
                channel = recording.channel_ids
                start_time = float(segment["start_time"])
                end_time = float(segment["end_time"])
                text = segment["text"]
                alignment = None

                if create_word_alignment:
                    alignment = {"word": []}

                    for alig_text, alig_start_time, alig_end_time in segment[
                        "word_timing"
                    ]:
                        # Skip all the fillings.
                        if "<" in alig_text or ">" in alig_text:
                            continue
                        alig_start_time = float(alig_start_time)
                        alig_end_time = float(alig_end_time)
                        alignment["word"].append(
                            AlignmentItem(
                                symbol=alig_text,
                                start=alig_start_time,
                                duration=alig_end_time - alig_start_time,
                            )
                        )

                supervisions.append(
                    SupervisionSegment(
                        id=f"{device_id}_{str(int(start_time*100)).zfill(6)}_{str(int(end_time*100)).zfill(6)}",
                        recording_id=recording.id,
                        start=start_time,
                        duration=end_time - start_time,
                        channel=channel,
                        text=text,
                        speaker=speaker_id,
                        alignment=alignment,
                    )
                )

            if is_multi_channel:
                mc_cuts.append(
                    MonoCut(
                        id=recording.id,
                        start=0,
                        duration=recording.duration,
                        channel=recording.channel_ids,
                        supervisions=supervisions,
                        recording=recording,
                    )
                )
            else:
                sc_cuts.append(
                    MonoCut(
                        id=recording.id,
                        start=0,
                        duration=recording.duration,
                        channel=0,
                        supervisions=supervisions,
                        recording=recording,
                    )
                )

    return sc_cuts, mc_cuts



import atexit
import contextlib
import ctypes
import ctypes.util
import os
from typing import Optional, Tuple

import numpy as np

__all__ = ["libsox", "libsox_rate", "libsox_available"]

libsox = None
sox_free = None
LIBSOX_INITIALIZED = False
SOX_SUCCESS = 0
SOX_ENCODING_FLOAT = 3


class sox_signalinfo_t(ctypes.Structure):
    _fields_ = [
        ("rate", ctypes.c_double),
        ("channels", ctypes.c_uint),
        ("precision", ctypes.c_uint),
        ("length", ctypes.c_uint64),
        ("mult", ctypes.POINTER(ctypes.c_double)),
    ]


class sox_encodinginfo_t(ctypes.Structure):
    _fields_ = [
        ("encoding", ctypes.c_uint),
        ("bits_per_sample", ctypes.c_uint),
        ("compression", ctypes.c_double),
        ("reverse_bytes", ctypes.c_uint),
        ("reverse_nibbles", ctypes.c_uint),
        ("reverse_bits", ctypes.c_uint),
        ("opposite_endian", ctypes.c_uint),
    ]


class sox_format_t(ctypes.Structure):
    _fields_ = [
        ("filename", ctypes.c_char_p),
        ("signal", sox_signalinfo_t),
        ("encoding", sox_encodinginfo_t),
        ("filetype", ctypes.c_char_p),
    ]


class sox_effect_handler_t(ctypes.Structure):
    pass


class sox_effects_chain_t(ctypes.Structure):
    pass


class sox_effect_t(ctypes.Structure):
    pass


def libsox_available() -> bool:
    return ctypes.util.find_library("sox") is not None


def libsox_import() -> None:
    global libsox
    global sox_free
    global LIBSOX_INITIALIZED
    if LIBSOX_INITIALIZED:
        return

    if not libsox_available():
        raise RuntimeError("libsox not available but import requested")

    libsox_ = ctypes.CDLL(ctypes.util.find_library("sox"))
    libc_ = ctypes.CDLL(ctypes.util.find_library("c"))

    sox_free_ = libc_.free
    sox_free_.argtypes = [ctypes.c_void_p]
    sox_free_.restype = None

    libsox_.sox_init.restype = ctypes.c_int
    libsox_.sox_quit.restype = ctypes.c_int

    libsox_.sox_open_mem_read.argtypes = [
        ctypes.c_void_p,
        ctypes.c_size_t,
        ctypes.POINTER(sox_signalinfo_t),
        ctypes.POINTER(sox_encodinginfo_t),
        ctypes.c_char_p,
    ]
    libsox_.sox_open_mem_read.restype = ctypes.POINTER(sox_format_t)

    libsox_.sox_open_mem_write.argtypes = [
        ctypes.c_void_p,
        ctypes.c_size_t,
        ctypes.POINTER(sox_signalinfo_t),
        ctypes.POINTER(sox_encodinginfo_t),
        ctypes.c_char_p,
        ctypes.c_char_p,
    ]
    libsox_.sox_open_mem_write.restype = ctypes.POINTER(sox_format_t)

    libsox_.sox_close.argtypes = [ctypes.POINTER(sox_format_t)]
    libsox_.sox_close.restype = ctypes.c_int

    libsox_.sox_create_effects_chain.argtypes = [
        ctypes.POINTER(sox_encodinginfo_t),
        ctypes.POINTER(sox_encodinginfo_t),
    ]
    libsox_.sox_create_effects_chain.restype = ctypes.POINTER(sox_effects_chain_t)

    libsox_.sox_delete_effects_chain.argtypes = [ctypes.POINTER(sox_effects_chain_t)]

    libsox_.sox_find_effect.argtypes = [ctypes.c_char_p]
    libsox_.sox_find_effect.restype = ctypes.POINTER(sox_effect_handler_t)

    libsox_.sox_create_effect.argtypes = [ctypes.POINTER(sox_effect_handler_t)]
    libsox_.sox_create_effect.restype = ctypes.POINTER(sox_effect_t)

    libsox_.sox_effect_options.argtypes = [
        ctypes.POINTER(sox_effect_t),
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_char_p),
    ]
    libsox_.sox_effect_options.restype = ctypes.c_int

    libsox_.sox_add_effect.argtypes = [
        ctypes.POINTER(sox_effects_chain_t),
        ctypes.POINTER(sox_effect_t),
        ctypes.POINTER(sox_signalinfo_t),
        ctypes.POINTER(sox_signalinfo_t),
    ]
    libsox_.sox_add_effect.restype = ctypes.c_int

    libsox_.sox_flow_effects.argtypes = [
        ctypes.POINTER(sox_effects_chain_t),
        ctypes.c_void_p,
        ctypes.c_void_p,
    ]
    libsox_.sox_flow_effects.restype = ctypes.c_int

    if libsox_.sox_init() != SOX_SUCCESS:
        raise RuntimeError("Failed to initialize SoX.")

    libsox = libsox_
    sox_free = sox_free_
    LIBSOX_INITIALIZED = True


def libsox_cleanup() -> None:
    global LIBSOX_INITIALIZED
    if not LIBSOX_INITIALIZED:
        return

    libsox.sox_quit()


atexit.register(libsox_cleanup)


def libsox_rate(
    audio: np.ndarray,
    sample_rate: int,
    target_rate: int,
    quality: str = "v",
    phase: str = "I",
    bandwidth: float = 99.0,
) -> Tuple[np.ndarray, int]:
    if not LIBSOX_INITIALIZED:
        libsox_import()

    if audio.ndim != 1 or audio.dtype != np.float32:
        raise ValueError("Audio must be a 1D float32 numpy array.")

    in_ft = None
    out_ft = None
    chain = None
    try:
        input_signal = sox_signalinfo_t(
            rate=float(sample_rate), channels=1, precision=32, length=len(audio)
        )
        input_encoding = sox_encodinginfo_t(
            encoding=SOX_ENCODING_FLOAT, bits_per_sample=32
        )
        in_ft = libsox.sox_open_mem_read(
            audio.ctypes.data_as(ctypes.c_void_p),
            audio.nbytes,
            ctypes.byref(input_signal),
            ctypes.byref(input_encoding),
            b"raw",
        )
        if not in_ft:
            raise RuntimeError("Failed to open memory for reading.")

        original_input_signal = sox_signalinfo_t()
        ctypes.memmove(
            ctypes.byref(original_input_signal),
            ctypes.byref(in_ft.contents.signal),
            ctypes.sizeof(sox_signalinfo_t),
        )

        max_output_samples = round(
            len(audio) * float(target_rate) / float(sample_rate) * 1.05
        )

        max_output_size = max_output_samples * 4
        output_buffer = (ctypes.c_char * max_output_size)()

        output_signal = sox_signalinfo_t(rate=target_rate, channels=1, precision=32)
        output_encoding = sox_encodinginfo_t(
            encoding=SOX_ENCODING_FLOAT, bits_per_sample=32
        )
        out_ft = libsox.sox_open_mem_write(
            output_buffer,
            max_output_size,
            ctypes.byref(output_signal),
            ctypes.byref(output_encoding),
            b"raw",
            None,
        )
        if not out_ft:
            raise RuntimeError("Failed to open memory for writing.")

        chain = libsox.sox_create_effects_chain(
            ctypes.byref(in_ft.contents.encoding),
            ctypes.byref(out_ft.contents.encoding),
        )
        if not chain:
            raise RuntimeError("Failed to create effects chain.")

        intermediate_signal = sox_signalinfo_t()
        ctypes.memmove(
            ctypes.byref(intermediate_signal),
            ctypes.byref(in_ft.contents.signal),
            ctypes.sizeof(sox_signalinfo_t),
        )

        e = libsox.sox_create_effect(libsox.sox_find_effect(b"input"))
        args = (ctypes.c_char_p * 1)(ctypes.cast(in_ft, ctypes.c_char_p))
        libsox.sox_effect_options(e, 1, args)
        libsox.sox_add_effect(
            chain,
            e,
            ctypes.byref(intermediate_signal),
            ctypes.byref(intermediate_signal),
        )
        sox_free(e)

        e = libsox.sox_create_effect(libsox.sox_find_effect(b"rate"))

        rate_args_list = [
            f"-{quality}",
            f"-{phase}",
            "-b",
            str(bandwidth),
            str(target_rate),
        ]

        num_rate_args = len(rate_args_list)
        args = (ctypes.c_char_p * num_rate_args)(
            *[s.encode("utf-8") for s in rate_args_list]
        )

        libsox.sox_effect_options(e, num_rate_args, args)
        libsox.sox_add_effect(
            chain,
            e,
            ctypes.byref(intermediate_signal),
            ctypes.byref(intermediate_signal),
        )
        sox_free(e)

        e = libsox.sox_create_effect(libsox.sox_find_effect(b"output"))
        args = (ctypes.c_char_p * 1)(ctypes.cast(out_ft, ctypes.c_char_p))
        libsox.sox_effect_options(e, 1, args)
        libsox.sox_add_effect(
            chain,
            e,
            ctypes.byref(intermediate_signal),
            ctypes.byref(out_ft.contents.signal),
        )
        sox_free(e)

        libsox.sox_flow_effects(chain, None, None)

    finally:
        if chain:
            libsox.sox_delete_effects_chain(chain)
        if out_ft:
            libsox.sox_close(out_ft)

        if in_ft:
            ctypes.memmove(
                ctypes.byref(in_ft.contents.signal),
                ctypes.byref(original_input_signal),
                ctypes.sizeof(sox_signalinfo_t),
            )
            libsox.sox_close(in_ft)

    num_output_samples = round(len(audio) * target_rate / sample_rate)

    # Ensure a standard, writeable numpy array is returned by making a copy.
    return np.frombuffer(
        output_buffer.raw[: num_output_samples * 4], dtype=np.float32
    ).copy(), int(target_rate)



import contextlib
import os
import typing
from typing import List, Literal, Optional, Union

ResamplingBackend = Literal["default", "sox"]
CURRENT_RESAMPLING_BACKEND: Optional[ResamplingBackend] = None


def set_current_resampling_backend(backend: ResamplingBackend) -> None:
    global CURRENT_RESAMPLING_BACKEND

    if backend not in available_resampling_backends():
        raise ValueError(
            f"Invalid resample backend: {backend}. Available backends: {available_resampling_backends()}"
        )

    CURRENT_RESAMPLING_BACKEND = backend


def get_current_resampling_backend() -> ResamplingBackend:
    global CURRENT_RESAMPLING_BACKEND

    if CURRENT_RESAMPLING_BACKEND is not None:
        return CURRENT_RESAMPLING_BACKEND

    maybe_env_backend = os.environ.get("LHOTSE_RESAMPLING_BACKEND")
    if maybe_env_backend:
        set_current_resampling_backend(maybe_env_backend)
        return CURRENT_RESAMPLING_BACKEND

    set_current_resampling_backend("default")
    return CURRENT_RESAMPLING_BACKEND


def available_resampling_backends() -> List[ResamplingBackend]:
    return list(typing.get_args(ResamplingBackend))


@contextlib.contextmanager
def resampling_backend(backend: Union[ResamplingBackend, str]):
    previous_backend = get_current_resampling_backend()
    set_current_resampling_backend(backend)
    yield
    set_current_resampling_backend(previous_backend)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.librimix import download_librimix, prepare_librimix
from lhotse.utils import Pathlike

__all__ = ["librimix"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("librispeech_root_path", type=click.Path(exists=True, dir_okay=True))
@click.argument("wham_recset_root_path", type=click.Path(exists=True, dir_okay=True))
@click.argument("librimix_metadata_path", type=click.Path(exists=True, dir_okay=True))
@click.argument("workdir", type=click.Path(exists=False, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-n",
    "--n_src",
    type=click.Choice(["2", "3"], case_sensitive=False),
    default="2",
    help="Number of sources used to create mixtures.",
)
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def librimix(
    librispeech_root_path: Pathlike,
    wham_recset_root_path: Pathlike,
    librimix_metadata_path: Pathlike,
    workdir: Pathlike,
    output_dir: Pathlike,
    n_src: str,
    num_jobs: int,
):
    """LibrMix source separation data preparation."""
    prepare_librimix(
        librispeech_root_path=librispeech_root_path,
        wham_recset_root_path=wham_recset_root_path,
        librimix_metadata_path=librimix_metadata_path,
        workdir=workdir,
        output_dir=output_dir,
        n_src=int(n_src),
        num_jobs=num_jobs,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def librimix(target_dir: Pathlike):
    """Mini LibriMix download."""
    download_librimix(target_dir)



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.librispeechmix import (
    download_librispeechmix,
    prepare_librispeechmix,
)
from lhotse.utils import Pathlike

__all__ = ["librispeechmix"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("librispeech_root_path", type=click.Path(exists=True, dir_okay=True))
@click.argument(
    "librispeechmix_metadata_path", type=click.Path(exists=True, dir_okay=True)
)
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use.",
)
def librispeechmix(
    librispeech_root_path: Pathlike,
    librispeechmix_metadata_path: Pathlike,
    output_dir: Pathlike,
    num_jobs: int,
):
    prepare_librispeechmix(
        librispeech_root_path=librispeech_root_path,
        librispeechmix_metadata_path=librispeechmix_metadata_path,
        output_dir=output_dir,
        num_jobs=num_jobs,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def librispeechmix(target_dir: Pathlike):
    download_librispeechmix(target_dir)



"""
WHAM noise recordings preparation for Lhotse.

This recipe prepares the noise component of the WHAM dataset, which consists
of real-world ambient noise recordings collected in various environments
(cafes, restaurants, bars, etc.). These noise recordings are commonly used
in combination with clean speech datasets to create noisy mixtures for
speech separation and enhancement tasks.

For more details about WHAM, see:
- Paper: "WHAM!: Extending Speech Separation to Noisy Environments"
  https://arxiv.org/abs/1907.01160
- Original dataset: https://wham.csail.mit.edu/
"""
import logging
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union
from zipfile import ZipFile

from lhotse import Recording, RecordingSet, SupervisionSet, validate
from lhotse.utils import Pathlike, resumable_download, safe_extract

WHAM_URL = (
    "https://my-bucket-a8b4b49c25c811ee9a7e8bba05fa24c7.s3.amazonaws.com/wham_noise.zip"
)


def download_wham(
    target_dir: Pathlike = ".",
    url: Optional[str] = WHAM_URL,
    force_download: Optional[bool] = False,
) -> Path:
    """
    Download and untar the WHAM corpus.

    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param url: str, the url that downloads file called "wham_noise.zip".
    :param force_download: bool, if True, download the archive even if it already exists.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    zip_name = "wham_noise.zip"
    zip_path = target_dir / zip_name
    corpus_dir = target_dir / "wham_noise"
    completed_detector = target_dir / ".wham_noise_completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {zip_name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(url, filename=zip_path, force_download=force_download)
    logging.info("Extracting files...")
    with ZipFile(zip_path) as zf:
        zf.extractall(path=target_dir)
        completed_detector.touch()
    return corpus_dir


def prepare_wham(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    manifests = {}
    splits = ["tr", "cv", "tt"]

    for split in splits:
        logging.info(f"Scanning {split} split...")
        manifests[split] = {"recordings": scan_recordings(corpus_dir / split)}
        logging.info(f"Validating {split} split...")
        validate(manifests[split]["recordings"])

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        for split in manifests:
            for key, manifest in manifests[split].items():
                manifest.to_file(output_dir / f"wham_{key}_{split}.jsonl.gz")

    return manifests


def scan_recordings(corpus_dir: Path) -> RecordingSet:
    return RecordingSet.from_recordings(
        Recording.from_file(file) for file in corpus_dir.rglob("*.wav")
    )



"""
LibriMix dataset preparation for Lhotse.

This recipe replicates the LibriMix dataset preparation by manipulating existing recordings
instead of generating and saving new audio files. LibriMix is an open source dataset for
source separation in noisy environments, derived from LibriSpeech signals (clean subset)
and WHAM noise.

The original dataset supports:
- Multiple sources (2 or 3 speakers) in mixtures
- Different sample rates (typically 16kHz and 8kHz)
- Different mixture modes:
  - min: mixture ends when the shortest source ends
  - max: mixture ends when the longest source ends
- Different mixture types:
  - mix_clean: utterances only
  - mix_both: utterances + noise
  - mix_single: 1 utterance + noise

**Current Limitations:**
This Lhotse recipe currently supports only:
- 16kHz sample rate
- 'max' mode (mixture ends when the longest source ends)

**Important Note on Quantization:**
The original LibriMix recipe introduces a quantization error when saving audio files via soundfile,
which by default uses PCM_16 format. If you need to replicate the exact quantization behavior from
the original recipe, you can apply the following transformation:

```python
import tempfile
import soundfile as sf

with tempfile.NamedTemporaryFile(suffix=".wav") as f:
    sf.write(f.name, cut.load_audio().T, 16000)
    audio_quantized, sr = sf.read(f.name)
```

Unlike the original LibriMix generation which creates ~430GB for Libri2Mix and ~332GB for Libri3Mix,
this recipe works with existing LibriSpeech and WHAM recordings and creates virtual mixtures,
making it much more storage efficient.

For more details about LibriMix, see:
- GitHub repository: https://github.com/JorisCos/LibriMix/
- Paper: "LibriMix: An Open-Source Dataset for Generalizable Speech Separation"
  https://arxiv.org/pdf/2005.11262.pdf

Citation:
@misc{cosentino2020librimix,
    title={LibriMix: An Open-Source Dataset for Generalizable Speech Separation},
    author={Joris Cosentino and Manuel Pariente and Samuele Cornell and Antoine Deleforge and Emmanuel Vincent},
    year={2020},
    eprint={2005.11262},
    archivePrefix={arXiv},
    primaryClass={eess.AS}
}
"""
import csv
import json
import logging
import os
import urllib.request
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np
import tqdm

import lhotse
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.audio.backend import info, save_audio
from lhotse.cut.set import CutSet, MonoCut, mix
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.utils import Pathlike

RATE = 16000


def _fetch_json(url):
    req = urllib.request.Request(url, headers={"User-Agent": "python-urllib"})
    with urllib.request.urlopen(req) as resp:
        return json.load(resp)


def _fetch_bytes(url):
    req = urllib.request.Request(url, headers={"User-Agent": "python-urllib"})
    with urllib.request.urlopen(req) as resp:
        return resp.read()


def download_github_dir(user, repo, path, branch="main", save_dir="."):
    api_url = f"https://api.github.com/repos/{user}/{repo}/contents/{path}?ref={branch}"
    files = _fetch_json(api_url)

    os.makedirs(save_dir, exist_ok=True)

    for file in files:
        file_path = os.path.join(save_dir, file["name"])
        if file["type"] == "file":
            with open(file_path, "wb") as f:
                f.write(_fetch_bytes(file["download_url"]))
        elif file["type"] == "dir":
            download_github_dir(user, repo, file["path"], branch, file_path)


def download_librimix(
    target_dir: Pathlike = ".",
) -> Path:
    """Download LibriMix metadata."""
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    metadata_dir = target_dir / "metadata"
    completed_detector = metadata_dir / ".completed"

    if completed_detector.is_file():
        logging.info(f"Skipping download because {completed_detector} exists.")
        return metadata_dir

    logging.info(
        f"Downloading https://github.com/JorisCos/LibriMix/tree/master/metadata to {metadata_dir}..."
    )
    os.makedirs(metadata_dir, exist_ok=True)
    download_github_dir("JorisCos", "LibriMix", "metadata", "master", metadata_dir)
    completed_detector.touch()
    return metadata_dir


def prepare_librimix(
    librispeech_root_path: Pathlike,
    wham_recset_root_path: Pathlike,
    librimix_metadata_path: Pathlike,
    workdir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    n_src: int = 2,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, CutSet]]:
    """
    Prepare LibriMix manifests for multi-speaker mixtures.

    Args:
        librispeech_root_path: Path to LibriSpeech manifests
        wham_recset_root_path: Path to WHAM noise manifests
        librimix_metadata_path: Path to LibriMix metadata
        output_dir: Directory to save manifests
        workdir: Working directory for temporary files
        n_src: Number of sources to for mixing
        num_jobs: Number of parallel threads used for processing (default: 1)

    Returns:
        Dict with keys for each split containing 'cuts' for both clean and noisy versions
    """
    logging.warning(
        "The original LibriMix recipe introduces a quantization error when saving audio files via soundfile, which by default uses PCM_16 format. If you need to replicate the exact quantization behavior from the original recipe, you can save and load the audio using a temporary file as shown in the docstring of this function."
    )
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    if workdir is not None:
        workdir = Path(workdir)
        workdir.mkdir(parents=True, exist_ok=True)

    manifests = {}

    # Collect all dataset parts
    dataset_parts = []

    n_src_meta_root = Path(librimix_metadata_path) / f"Libri{n_src}Mix"
    md_filename_list = [
        file
        for file in os.listdir(n_src_meta_root)
        if "info" not in file and file != ".completed"
    ]
    for md_filename in md_filename_list:
        part_name = f"{md_filename.replace('.csv', '')}_clean"
        part_name_noisy = f"{md_filename.replace('.csv', '')}_noisy"
        dataset_parts.extend([part_name, part_name_noisy])

    # Maybe the manifests already exist: we can read them and save a bit of preparation time.
    if output_dir is not None:
        manifests = read_manifests_if_cached(
            dataset_parts=dataset_parts,
            output_dir=output_dir,
            prefix="librimix",
            types=("cutset",),
        )

    # Load WHAM recordings with speed augmentation
    wham_recsets = _load_wham_recordings(wham_recset_root_path)

    n_src_meta_root = Path(librimix_metadata_path) / f"Libri{n_src}Mix"
    md_filename_list = [
        file
        for file in os.listdir(n_src_meta_root)
        if "info" not in file and file != ".completed"
    ]

    for md_filename in md_filename_list:
        part_name = f"{md_filename.replace('.csv', '')}"
        part_name_noisy = f"{md_filename.replace('.csv', '')}_noisy"

        # Check if manifests already exist and are cached
        if manifests_exist(
            part=part_name, output_dir=output_dir, prefix="librimix", types=("cutset",)
        ) and manifests_exist(
            part=part_name_noisy,
            output_dir=output_dir,
            prefix="librimix",
            types=("cutset",),
        ):
            logging.info(
                f"LibriMix subset: {part_name} and {part_name_noisy} already prepared - skipping."
            )
            continue

        clean_cuts, noisy_cuts = _process_metadata_file(
            md_filename,
            n_src_meta_root,
            n_src,
            librispeech_root_path,
            wham_recsets,
            workdir,
            num_jobs,
        )

        # As we need to keep MixedCuts together we cannot decompose to rec- and supsets.
        # Process clean version
        clean_cutset = CutSet.from_cuts(clean_cuts)

        if output_dir is not None:
            clean_cutset.to_file(output_dir / f"librimix_cutset_{part_name}.jsonl.gz")

        manifests[part_name] = {
            "cutset": clean_cutset,
        }

        noisy_cutset = CutSet.from_cuts(noisy_cuts)

        if output_dir is not None:
            noisy_cutset.to_file(
                output_dir / f"librimix_cutset_{part_name_noisy}.jsonl.gz"
            )

        manifests[part_name_noisy] = {
            "cutset": noisy_cutset,
        }

    return manifests


def _load_wham_recordings(wham_recset_root_path: Pathlike) -> Dict[str, RecordingSet]:
    """Load WHAM recordings with speed augmentation for training set."""
    wham_splits = [
        ("train", "wham_recordings_tr.jsonl.gz"),
        ("dev", "wham_recordings_cv.jsonl.gz"),
        ("test", "wham_recordings_tt.jsonl.gz"),
    ]
    speed_factors = [0.8, 1.0, 1.2]

    wham_recsets = {
        key: lhotse.load_manifest(Path(wham_recset_root_path) / split)
        for key, split in wham_splits
    }
    wham_recsets["train"] = _augment_wham(wham_recsets["train"], speed_factors)
    return wham_recsets


def _augment_wham(recset: RecordingSet, speed_factors: List[float]) -> RecordingSet:
    """Apply speed augmentation to WHAM recordings."""
    import re

    def fix_rec_ids(recording: Recording) -> Recording:
        recording.id = re.sub(r"_sp(\d+)\.(\d+)$", r"sp\1\2", recording.id)
        return recording

    new_recset = []
    for speed_factor in speed_factors:
        if speed_factor != 1.0:
            augmented_recset = recset.perturb_speed(speed_factor)
        else:
            augmented_recset = recset
        augmented_recset = augmented_recset.map(fix_rec_ids)
        new_recset.extend(augmented_recset)
    return RecordingSet.from_recordings(new_recset)


def _extend_noise(noise: np.ndarray, max_length: int) -> np.ndarray:
    """Concatenate noise using Hanning window."""
    noise_ex = noise
    window = np.hanning(RATE + 1)
    i_w = window[: len(window) // 2 + 1]
    d_w = window[len(window) // 2 :: -1]

    while len(noise_ex) < max_length:
        noise_ex = np.concatenate(
            (
                noise_ex[: len(noise_ex) - len(d_w)],
                np.multiply(noise_ex[len(noise_ex) - len(d_w) :], d_w)
                + np.multiply(noise[: len(i_w)], i_w),
                noise[len(i_w) :],
            )
        )
    return noise_ex[:max_length]


def _process_row(
    row: dict,
    librispeech_cutset: CutSet,
    wham_recset: RecordingSet,
    n_src: int,
    workdir: Optional[Path],
):
    """Process a single row from metadata CSV."""
    # Extract source information
    srcs = []
    gains = []
    for src in range(1, n_src + 1):
        srcs.append(Path(row[f"source_{src}_path"]).stem)
        gains.append(row[f"source_{src}_gain"])

    srcs = librispeech_cutset.subset(cut_ids=srcs)
    normalized_cuts = [src.perturb_volume(gain) for src, gain in zip(srcs, gains)]

    # Build clean mix
    clean_mix = normalized_cuts.pop()
    while normalized_cuts:
        clean_mix = mix(clean_mix, normalized_cuts.pop())
    clean_mix.id = row["mixture_ID"]

    # Process noise
    noise_id = Path(row["noise_path"]).stem
    noise_rec = wham_recset[noise_id]
    noise_gain = row[f"noise_gain"]
    noise_rec_perturbed = noise_rec.perturb_volume(noise_gain)

    if noise_rec_perturbed.duration < clean_mix.duration:
        noise_rec_perturbed = _extend_noise_recording(
            noise_rec_perturbed, clean_mix, row["mixture_ID"], workdir
        )

    noise_cut = MonoCut(
        id="noise",
        start=0,
        duration=clean_mix.duration,
        channel=0,
        recording=noise_rec_perturbed,
    )
    noisy_mix = mix(clean_mix, noise_cut, preserve_id="left")

    return clean_mix, noisy_mix


def _extend_noise_recording(
    noise_rec: Recording, clean_mix: MonoCut, mixture_id: str, workdir: Optional[Path]
) -> Recording:
    """Extend noise recording to match clean mix duration."""
    if workdir is None:
        workdir = Path(".")

    save_to = workdir / f"{noise_rec.id}_{mixture_id}.wav"
    if not save_to.exists():
        noise_array = noise_rec.load_audio()
        if noise_array.ndim > 1:
            noise_array = noise_array[0]

        extended_noise = _extend_noise(
            noise_array, int(clean_mix.duration * clean_mix.sampling_rate)
        )
        save_audio(
            dest=save_to, src=extended_noise, sampling_rate=noise_rec.sampling_rate
        )

    noise_rec_info = info(save_to)
    new_source = AudioSource(type="file", channels=[0], source=str(save_to))

    return Recording(
        id=noise_rec.id,
        sources=[new_source],
        sampling_rate=noise_rec_info.samplerate,
        num_samples=noise_rec_info.frames,
        duration=noise_rec_info.duration,
    )


def _read_metadata_csv(csv_path: Path) -> List[dict]:
    """
    Read LibriMix metadata using Python's standard csv library and cast gain fields to float.
    """
    rows: List[dict] = []
    with open(csv_path, "r", newline="") as f:
        reader = csv.DictReader(f)
        for row in reader:
            # Normalize/convert values as needed
            for k, v in list(row.items()):
                if k.endswith("_gain"):
                    row[k] = float(v)
            rows.append(row)
    return rows


def _process_metadata_file(
    md_filename: str,
    n_src_meta_root: Path,
    n_src: int,
    librispeech_root_path: Path,
    wham_recsets: Dict[str, RecordingSet],
    workdir: Optional[Path],
    num_jobs: int,
) -> Tuple[List[MonoCut], List[MonoCut]]:
    """Process a single metadata file and return clean and noisy cuts."""
    csv_path = n_src_meta_root / md_filename

    rows = _read_metadata_csv(csv_path)

    librispeech_cutset = lhotse.load_manifest(
        Path(librispeech_root_path)
        / md_filename.replace(f"libri{n_src}mix", "librispeech_cutset").replace(
            ".csv", ".jsonl.gz"
        )
    )
    librispeech_cutset = librispeech_cutset.modify_ids(
        lambda c: "-".join(c.split("-")[:-1])
    )

    split_name = "".join(md_filename.split("_")[1:]).split("-")[0]
    wham_recset = wham_recsets[split_name]

    clean_cuts = []
    noisy_cuts = []

    logging.info(f"Processing {md_filename}...")
    with ThreadPoolExecutor(max_workers=num_jobs) as ex:
        futures = [
            ex.submit(
                _process_row, row, librispeech_cutset, wham_recset, n_src, workdir
            )
            for row in rows
        ]
        for f in tqdm.tqdm(as_completed(futures), total=len(futures)):
            clean_mix, noisy_mix = f.result()
            clean_cuts.append(clean_mix)
            noisy_cuts.append(noisy_mix)

    return clean_cuts, noisy_cuts



import io
import logging
import typing
from dataclasses import dataclass
from typing import ClassVar, Literal, Optional, Tuple

import numpy as np

from lhotse.augmentation.transform import AudioTransform

Codec = Literal["opus", "mp3", "vorbis", "gsm"]

OPUS_SUPPORTED_SAMPLING_RATES = [8000, 12000, 16000, 24000, 48000]
MP3_SUPPORTED_SAMPLING_RATES = [
    8000,
    11025,
    12000,
    16000,
    22050,
    24000,
    32000,
    44100,
    48000,
]


@dataclass
class Compress(AudioTransform):
    """
    Modifies audio by running it through a lossy codec.

    :param codec: Used lossy audio codec. One of ``"opus"``, ``"mp3"``, ``"vorbis"``, or ``"gsm"``.
    :param compression_level: The level of compression to apply. 0.0 is for the lowest amount of compression, 1.0 is for highest. Ignored for ``"gsm"``.
    :return: The modified audio samples.
    """

    supported_codecs: ClassVar[Tuple[Codec]] = tuple(typing.get_args(Codec))
    codec: Codec
    compression_level: Optional[float] = None

    def __post_init__(self):
        if self.codec not in self.supported_codecs:
            raise ValueError(f"Unsupported augmentation codec {self.codec}")
        if not 0 <= self.compression_level <= 1:
            raise ValueError("Compression level must be between 0 and 1")

    def __call__(
        self,
        samples: np.ndarray,
        sampling_rate: int,
    ):
        # argument `sample_rate` is the original sampling rate of the `Recording` before any transforms are applied
        # even if we Resample prior to Compress, it is useless to check for it here
        # we just assume 8 kHz for GSM and original sampling rate for other codecs

        import soundfile as sf

        # lhotse: (channels, samples)
        # soundfile: (samples, channels)

        channels, _ = samples.shape
        samples = samples.transpose(1, 0)

        with io.BytesIO() as buffer:
            sf.write(
                buffer,
                samples,
                samplerate=sampling_rate if self.codec != "gsm" else 8000,
                closefd=False,
                **self.prepare_sf_arguments(),
            )
            data = buffer.getvalue()
        with io.BytesIO(data) as f:
            if self.codec == "gsm":
                samples_compressed, sampling_rate_compressed = sf.read(
                    f,
                    always_2d=True,
                    samplerate=sampling_rate if self.codec != "gsm" else 8000,
                    channels=channels,
                    format="RAW",
                    subtype="GSM610",
                    dtype=np.float32,
                )
            else:
                samples_compressed, sampling_rate_compressed = sf.read(
                    f, always_2d=True, dtype=np.float32
                )

        # when one writes Opus files with soundfile,
        # it adds extra information in the file header about the original sampling rate,
        # so that when we load this file with soundfile later,
        # it's resampled to original sampling rate
        # before returning the audio array (unlike most other tools)

        samples_compressed = samples_compressed.transpose(1, 0)

        return samples_compressed

    def prepare_sf_arguments(self) -> dict:
        import soundfile as sf

        if self.codec == "mp3":
            return {
                "compression_level": self.compression_level,
                "format": "MP3",
                "subtype": sf.default_subtype("MP3"),
            }
        elif self.codec == "opus":
            return {
                "compression_level": self.compression_level,
                "format": "OGG",
                "subtype": "OPUS",
            }
        elif self.codec == "vorbis":
            return {
                "compression_level": self.compression_level,
                "format": "OGG",
                "subtype": "VORBIS",
            }
        elif self.codec == "gsm":
            return {"format": "RAW", "subtype": "GSM610"}
        else:
            raise NotImplementedError(f"Unsupported augmentation codec {self.codec}")

    def reverse_timestamps(self, offset, duration, sampling_rate):
        return offset, duration



from typing import Dict, List, Optional, Tuple, Union

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.bengaliai_speech import prepare_bengaliai_speech
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def bengaliai_speech(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
):
    """Bengali.AI Speech data preparation."""
    prepare_bengaliai_speech(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
        num_jobs=num_jobs,
    )



"""
The Grid Corpus is a large multitalker audiovisual sentence corpus designed to support joint
computational-behavioral studies in speech perception. In brief, the corpus consists of high-quality
audio and video (facial) recordings of 1000 sentences spoken by each of 34 talkers (18 male, 16 female),
for a total of 34000 sentences. Sentences are of the form "put red at G9 now".

Source: https://zenodo.org/record/3625687
"""
import os
import shutil
import subprocess
import tempfile
import zipfile
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import (
    Recording,
    RecordingSet,
    SupervisionSet,
    fix_manifests,
    validate_recordings_and_supervisions,
)
from lhotse.supervision import AlignmentItem, SupervisionSegment
from lhotse.utils import Pathlike, is_module_available

GRID_ZENODO_ID = "10.5281/zenodo.3625687"


def download_grid(
    target_dir: Pathlike = ".",
    force_download: bool = False,
) -> Path:
    """
    Download and untar the dataset, supporting both LibriSpeech and MiniLibrispeech

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: "librispeech", "mini_librispeech",
        or a list of splits (e.g. "dev-clean") to download.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param alignments: should we download the alignments. The original source is:
        https://github.com/CorentinJ/librispeech-alignments
    :param base_url: str, the url of the OpenSLR resources.
    :param alignments_url: str, the url of LibriSpeech word alignments
    :return: the path to downloaded and extracted directory with data.
    """
    if not is_module_available("zenodo_get"):
        raise RuntimeError(
            "To download Grid Audio-Visual Speech Corpus please 'pip install zenodo_get'."
        )

    corpus_dir = Path(target_dir)
    corpus_dir.mkdir(parents=True, exist_ok=True)

    download_marker = corpus_dir / ".downloaded"
    if not download_marker.exists() or force_download:
        subprocess.run(
            f"zenodo_get {GRID_ZENODO_ID}", shell=True, check=True, cwd=corpus_dir
        )
        download_marker.touch()

    for p in tqdm(corpus_dir.glob("*.zip"), desc="Unzipping files"):
        with zipfile.ZipFile(p) as f:
            f.extractall(corpus_dir)

    # Speaker mapping to fix mis-assigned alignment data
    speaker_fix_map = {
        "s1": "s1",
        "s2": "s2",
        "s3": "s3",
        "s4": "s4",
        "s5": "s6",
        "s6": "s5",
        "s7": "s7",
        "s8": "s8",
        "s9": "s9",
        "s10": "s13",
        "s11": "s10",
        "s12": "s11",
        "s13": "s12",
        "s14": "s15",
        "s15": "s14",
        "s16": "s16",
        "s17": "s17",
        "s18": "s19",
        "s19": "s18",
        "s20": "s21",
        "s22": "s23",
        "s23": "s22",
        "s24": "s24",
        "s25": "s25",
        "s26": "s27",
        "s27": "s26",
        "s28": "s29",
        "s29": "s28",
        "s30": "s30",
        "s31": "s31",
        "s32": "s33",
        "s33": "s32",
        "s34": "s34",
    }

    # Downloaded alignment folder has mis-assigned speaker folders, we fix it here
    input_dir = corpus_dir / "alignments"
    tempfile.tempdir = os.path.abspath(corpus_dir)
    temp_alignment_dir = tempfile.mkdtemp()

    for tgt_folder, src_folder in speaker_fix_map.items():
        src_path = os.path.join(input_dir, src_folder)
        tgt_path = os.path.join(temp_alignment_dir, tgt_folder)
        shutil.copytree(src_path, tgt_path)
        print(f"Copied entire folder from {src_folder} to {tgt_folder}")

    shutil.rmtree(input_dir)
    os.rename(temp_alignment_dir, input_dir)

    return corpus_dir


def prepare_grid(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    with_supervisions: bool = True,
    num_jobs: int = 1,
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param with_supervisions: bool, when False, we'll only return recordings; when True, we'll also
        return supervisions created from alignments, but might remove some recordings for which
        they are missing.
    :param num_jobs: int, number of parallel jobs.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    ali_dir = corpus_dir / "alignments"
    assert ali_dir.is_dir()
    audio_dir = corpus_dir / "audio_25k"
    assert audio_dir.is_dir()

    recordings = []
    supervisions = []

    futures = []
    # gather all .mpg files in the corpus directory
    all_mpg_files = list(Path(corpus_dir).rglob("*.mpg"))
    all_mpg_files = [f for f in all_mpg_files if "MACOSX" not in str(f)]

    with ProcessPoolExecutor(num_jobs) as ex:

        for video_path in all_mpg_files:
            speaker = video_path.parent.name
            futures.append(
                ex.submit(
                    process_single, video_path, speaker, ali_dir, with_supervisions
                )
            )

        for f in tqdm(
            as_completed(futures), total=len(futures), desc="Scanning videos"
        ):
            try:
                result = f.result()
                if result is None:
                    continue
            except Exception as e:
                continue

            recording, maybe_supervision = result
            recordings.append(recording)
            if maybe_supervision is not None:
                supervisions.append(maybe_supervision)

    recordings = RecordingSet.from_recordings(recordings)
    if with_supervisions:
        supervisions = SupervisionSet.from_segments(supervisions)
        recordings, supervisions = fix_manifests(recordings, supervisions)
        validate_recordings_and_supervisions(recordings, supervisions)

    if output_dir is not None:
        output_dir = Path(output_dir)
        recordings.to_file(output_dir / "grid_recordings.jsonl.gz")
        if with_supervisions:
            supervisions.to_file(output_dir / "grid_supervisions.jsonl.gz")

    ans = {"recordings": recordings}
    if with_supervisions:
        ans.update(supervisions=supervisions)

    return ans


def process_single(
    video_path: Path, speaker: str, ali_dir: Path, with_supervisions: bool
):
    video_id = video_path.stem
    try:
        recording = Recording.from_file(
            video_path, recording_id=f"{speaker}_{video_id}"
        )
    except Exception as e:
        print(f"Unexpected error for {video_path}: {e}")
        return None

    supervision = None
    ali_path = (ali_dir / speaker / video_id).with_suffix(".align")
    if with_supervisions and ali_path.is_file():
        ali = [
            AlignmentItem(
                symbol=w,
                start=float(b) / 1000,
                duration=float(int(e) - int(b)) / 1000,
            )
            for b, e, w in (line.split() for line in ali_path.read_text().splitlines())
        ]
        supervision = SupervisionSegment(
            id=recording.id,
            recording_id=recording.id,
            start=0.0,
            duration=recording.duration,
            channel=recording.channel_ids,
            text=" ".join(item.symbol for item in ali if item.symbol != "sil"),
            language="English",
            speaker=speaker,
            alignment={"word": ali},
        )

    return recording, supervision



import logging
import os
import sys
from pathlib import Path


def default_tools_cachedir(force_mkdir: bool = False) -> Path:
    d = Path.home() / ".lhotse/tools"
    try:
        d.mkdir(exist_ok=True, parents=True)
    except OSError:
        if force_mkdir:
            raise
        else:
            logging.warning(
                f"We couldn't create lhotse utilities directory: {d} (not enough space/no permissions?)"
            )
    return d


def add_tools_to_path():
    sph2pipe_path = str(default_tools_cachedir() / "sph2pipe-2.5")
    sys.path.append(sph2pipe_path)
    os.environ["PATH"] += os.pathsep + sph2pipe_path  # platform-agnostic


def add_macos_homebrew_lib_paths():
    if os.uname().sysname == "Darwin":  # macOS
        HOMEBREW_LIB_PATHS = ["/opt/homebrew/lib", "/usr/local/lib"]
        for path in HOMEBREW_LIB_PATHS:
            dyld_library_path = os.environ.get("DYLD_LIBRARY_PATH", "")
            if os.path.exists(path) and path not in dyld_library_path.split(":"):
                os.environ["DYLD_LIBRARY_PATH"] = dyld_library_path + f":{path}"



import random
from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple, Union

from lhotse import CutSet
from lhotse.augmentation.compress import Codec
from lhotse.dataset.dataloading import resolve_seed


@dataclass
class Compress:
    """
    Applies a lossy compression algorithm filter to each Cut in a CutSet. The audio is decompressed back to raw waveforms.

    The compression is applied with a probability of ``p``. The codec is
    randomly selected the list of provided codecs,
    with optional weights controlling the selection.
    If compression level is provided as an interval,
    then the actual value is sampled uniformly from the provided interval.

    :param codecs: A list of codecs (supported: "opus", "mp3", "vorbis", "gsm")
    :param compression_level: A single value or an interval. 0.0 = lowest compression (highest bitrate), 1.0 = highest compression (lowest bitrate). If an interval is provided, the value is sampled uniformly.
    :param codec_weights: Optional weights for each codec (default: equal weights).
    :param p: The probability of applying the low-pass filter (default: 0.5).
    :param randgen: An optional random number generator (default: a new instance).
    :param preserve_id: Whether to preserve the original cut ID (default: False).
    """

    codecs: List[Codec]
    compression_level: Union[float, Tuple[float, float]] = 0.9
    codec_weights: Optional[List[float]] = None
    compress_custom_fields: bool = False
    p: float = 0.5
    seed: Union[int, Literal["trng", "randomized"]] = 42
    rng: Optional[random.Random] = None
    preserve_id: bool = False

    def __post_init__(self) -> None:
        assert sorted(self.codecs) == sorted(list(set(self.codecs))), "duplicate codecs"

        if isinstance(self.compression_level, (Tuple, List)):
            assert (
                len(self.compression_level) == 2
            ), f"Expected compression_level to be a tuple or a list with two values, got {self.compression_level}"
            min_compression, max_compression = self.compression_level
            assert (
                min_compression < max_compression
            ), f"Expected min_compression < max_compression, got {min_compression} >= {max_compression}"

        assert 0 <= self.p <= 1, f"Probability p must be between 0 and 1, got {self.p}"

        if self.codec_weights:
            assert len(self.codec_weights) == len(
                self.codecs
            ), f"Expected codec_weights to be a list with the same length as codecs, got len({self.codec_weights}) != len({self.codecs})"
            assert all(
                w >= 0 for w in self.codec_weights
            ), "All codec weights must be non-negative"
        else:
            # all codecs have equal weights by default
            self.codec_weights = [1.0 for _ in self.codecs]

        if self.rng is not None and self.seed is not None:
            raise ValueError("Either rng or seed must be provided, not both")
        if self.rng is None:
            self.rng = random.Random(resolve_seed(self.seed))

    def __call__(self, cuts: CutSet) -> CutSet:
        compressed_cuts = []
        for cut in cuts:
            if self.rng.random() <= self.p:
                if isinstance(self.compression_level, (Tuple, List)):
                    min_compression, max_compression = self.compression_level
                    compression_level = (
                        self.rng.random() * (max_compression - min_compression)
                        + min_compression
                    )
                else:
                    compression_level = self.compression_level

                codec, *_ = self.rng.choices(self.codecs, weights=self.codec_weights)
                new_cut = cut.compress(
                    codec=codec,
                    compression_level=compression_level,
                    compress_custom_fields=self.compress_custom_fields,
                )
                if not self.preserve_id:
                    new_cut.id = f"{new_cut.id}_{codec}_{compression_level:.2f}"
                compressed_cuts.append(new_cut)
            else:
                compressed_cuts.append(cut)

        return CutSet(compressed_cuts)



from typing import Any, Dict, Iterator, Optional, Tuple

from lhotse.array import Array, TemporalArray
from lhotse.audio.recording import Recording
from lhotse.cut import CutSet
from lhotse.features.base import Features
from lhotse.image import Image
from lhotse.serialization import get_aistore_client
from lhotse.utils import is_module_available, is_valid_url

# Mapping between Lhotse file storage types and in-memory equivalents.
FILE_TO_MEMORY_TYPE = {
    "numpy_files": "memory_raw",
    "lilcom_files": "memory_lilcom",
    "pillow_files": "memory_pillow",
}

ARCHIVE_EXTENSIONS = (".tar.gz", ".tar", ".tgz")


class AISBatchLoaderError(Exception):
    """Base exception for AISBatchLoader operations."""


class AISBatchLoader:
    """
    Loads all data referenced by a :class:`CutSet` in a single AIStore Get-Batch call.

    The loader optimizes I/O by aggregating all object URLs from a CutSet and requesting
    them together. It offloads archive extraction and data slicing to AIStore, avoiding
    redundant downloads and local decompression.

    Example:
        >>> loader = AISBatchLoader()
        >>> cuts_with_data = loader(cuts)
    """

    def __init__(self) -> None:
        """Initialize the AISBatchLoader with an AIStore client and batch context."""
        if not is_module_available("aistore"):
            raise ImportError(
                "Please run 'pip install aistore>=1.17.0' to use AISBatchLoader."
            )
        self.client, _ = get_aistore_client()

    def __call__(self, cuts: CutSet) -> CutSet:
        """
        Fetch all data referenced by a CutSet in one AIStore batch operation.

        Args:
            cuts: A non-lazy CutSet representing a single batch of data.

        Returns:
            The same CutSet object with all manifests updated to reference in-memory data.

        Raises:
            ValueError: If the input CutSet is lazy.
            AISBatchLoaderError: For invalid URLs or unsupported storage types.
        """
        if cuts.is_lazy:
            raise ValueError(
                "Lazy CutSets cannot be used with AISBatchLoader. "
                "Convert to eager via `cuts.to_eager()` before loading."
            )

        batch = self.client.batch()
        # Collect all URLs for get-batch and track which manifests have URLs
        manifest_list = []
        for cut in cuts:
            for _, manifest in cut.iter_data():
                has_url = self._collect_manifest_urls(manifest, batch)
                manifest_list.append((manifest, has_url))

        # Execute batch request
        batch_result = batch.get()

        # Apply the received data back into each manifest that had a URL
        for manifest, has_url in manifest_list:
            if has_url:
                _, content = next(batch_result)
                self._inject_data_into_manifest(manifest, content)

        return cuts

    # ----------------------------- Internal Helpers -----------------------------

    def _collect_manifest_urls(self, manifest: Any, batch: Any) -> None:
        """Add all URLs referenced in a manifest to the batch."""
        if isinstance(manifest, Recording):
            for source in manifest.sources:
                if source.type == "url":
                    self._add_url_to_batch(source.source, batch)
                    return True
            return False

        elif isinstance(manifest, TemporalArray):
            # TemporalArray wraps an Array, so we need to access the inner array
            inner_array = manifest.array
            if inner_array.storage_type not in FILE_TO_MEMORY_TYPE:
                raise AISBatchLoaderError(
                    f"Unsupported storage type '{inner_array.storage_type}'. "
                    f"Supported types: {list(FILE_TO_MEMORY_TYPE.keys())}"
                )

            obj_path = f"{inner_array.storage_path}/{inner_array.storage_key}"
            if is_valid_url(obj_path):
                self._add_url_to_batch(obj_path, batch)
                return True
            return False

        elif isinstance(manifest, (Array, Features, Image)):
            if manifest.storage_type not in FILE_TO_MEMORY_TYPE:
                raise AISBatchLoaderError(
                    f"Unsupported storage type '{manifest.storage_type}'. "
                    f"Supported types: {list(FILE_TO_MEMORY_TYPE.keys())}"
                )

            obj_path = f"{manifest.storage_path}/{manifest.storage_key}"
            if is_valid_url(obj_path):
                self._add_url_to_batch(obj_path, batch)
                return True
            return False

        return False

    def _add_url_to_batch(self, url: str, batch: Any) -> None:
        """Add a single AIStore URL to the batch request."""
        from aistore.sdk.utils import parse_url

        provider, bck_name, obj_name = parse_url(url)
        if not (provider and bck_name and obj_name):
            raise AISBatchLoaderError(f"Invalid object URL: '{url}'")

        arch_ext = self._get_archive_extension(obj_name)
        archpath = None
        if arch_ext and arch_ext in obj_name:
            prefix, _, suffix = obj_name.partition(f"{arch_ext}/")
            obj_name, archpath = prefix + arch_ext, suffix

        bucket = self.client.bucket(bck_name, provider)
        batch.add(bucket.object(obj_name), archpath=archpath)

    def _inject_data_into_manifest(self, manifest: Any, content: bytes) -> None:
        """Replace manifest storage references with in-memory content."""
        if isinstance(manifest, Recording):
            for source in manifest.sources:
                source.type = "memory"
                source.source = content

        elif isinstance(manifest, TemporalArray):
            # TemporalArray wraps an Array, so update the inner array
            inner_array = manifest.array
            inner_array.storage_type = FILE_TO_MEMORY_TYPE[inner_array.storage_type]
            inner_array.storage_path = ""
            inner_array.storage_key = content

        elif isinstance(manifest, (Array, Features, Image)):
            manifest.storage_type = FILE_TO_MEMORY_TYPE[manifest.storage_type]
            manifest.storage_path = ""
            manifest.storage_key = content

    @staticmethod
    def _get_archive_extension(obj_name: str) -> Optional[str]:
        """Return the supported archive extension if present in the object name."""
        for ext in ARCHIVE_EXTENSIONS:
            if ext in obj_name:
                return ext
        return None



import math
import random
from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple, Union

from lhotse import CutSet
from lhotse.dataset.dataloading import resolve_seed


@dataclass
class LowpassUsingResampling:
    """
    Applies a low-pass filter to each Cut in a CutSet by resampling the audio back and forth.
    """

    p: float = 0.5
    frequencies_interval: Tuple[float, float] = (3500, 8000)
    seed: Union[int, Literal["trng", "randomized"]] = 42
    rng: Optional[random.Random] = None
    preserve_id: bool = False

    def __post_init__(self) -> None:
        if self.rng is not None and self.seed is not None:
            raise ValueError("Either rng or seed must be provided, not both")
        if self.rng is None:
            self.rng = random.Random(resolve_seed(self.seed))

    def __call__(self, cuts: CutSet) -> CutSet:
        lowpassed_cuts = []
        for cut in cuts:
            if self.rng.random() <= self.p:
                low, high = self.frequencies_interval
                if high > cut.sampling_rate // 2:
                    raise ValueError(
                        f"Upper frequency limit {high} is greater than sampling rate / 2 ({cut.sampling_rate // 2})"
                    )

                # sampling from log-uniform[low, high] distribution
                cutoff_frequency = math.exp(
                    self.rng.uniform(math.log(low), math.log(high))
                )
                cutoff_frequency = int(cutoff_frequency)

                new_cut = cut.resample(cutoff_frequency * 2).resample(cut.sampling_rate)
                if not self.preserve_id:
                    new_cut.id = f"{cut.id}_lowpassed{cutoff_frequency:.0f}"
                lowpassed_cuts.append(new_cut)
            else:
                lowpassed_cuts.append(cut)

        return CutSet(lowpassed_cuts)



"""
LibriSpeechMix dataset preparation for Lhotse.

This recipe processes LibriSpeechMix metadata to create multi-speaker mixtures
from LibriSpeech recordings. LibriSpeechMix provides predefined
speaker combinations and timing information for creating consistent mixtures.

For more details, see:
- GitHub repository: https://github.com/NaoyukiKanda/LibriSpeechMix/
"""
import glob
import json
import logging
import os
import urllib.request
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, List, Optional

import tqdm

from lhotse.cut.set import CutSet, MonoCut, mix
from lhotse.recipes.utils import manifests_exist, read_manifests_if_cached
from lhotse.utils import Pathlike

RATE = 16000


def _fetch_json(url):
    req = urllib.request.Request(url, headers={"User-Agent": "python-urllib"})
    with urllib.request.urlopen(req) as resp:
        return json.load(resp)


def _fetch_bytes(url):
    req = urllib.request.Request(url, headers={"User-Agent": "python-urllib"})
    with urllib.request.urlopen(req) as resp:
        return resp.read()


def download_github_dir(user, repo, path, branch="main", save_dir="."):
    api_url = f"https://api.github.com/repos/{user}/{repo}/contents/{path}?ref={branch}"
    files = _fetch_json(api_url)

    os.makedirs(save_dir, exist_ok=True)

    for file in files:
        file_path = os.path.join(save_dir, file["name"])
        if file["type"] == "file":
            with open(file_path, "wb") as f:
                f.write(_fetch_bytes(file["download_url"]))
        elif file["type"] == "dir":
            download_github_dir(user, repo, file["path"], branch, file_path)


def download_librispeechmix(
    target_dir: Pathlike = ".",
) -> Path:
    """Download LibriSpeechMix metadata."""
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    metadata_dir = target_dir / "list"
    completed_detector = metadata_dir / ".completed"

    if completed_detector.is_file():
        logging.info(f"Skipping download because {completed_detector} exists.")
        return metadata_dir

    logging.info(
        f"Downloading https://github.com/NaoyukiKanda/LibriSpeechMix/tree/main/list to {metadata_dir}..."
    )
    download_github_dir("NaoyukiKanda", "LibriSpeechMix", "list", "main", metadata_dir)
    completed_detector.touch()
    return metadata_dir


def prepare_librispeechmix(
    librispeech_root_path: Pathlike,
    librispeechmix_metadata_path: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: Optional[int] = 1,
) -> Dict[str, Dict[str, CutSet]]:
    """
    Prepare LibriSpeechMix manifests for multi-speaker mixtures.

    Args:
        librispeech_root_path: Path to LibriSpeech manifests
        librispeechmix_metadata_path: Path to LibriSpeechMix metadata JSONL files
        output_dir: Directory to save manifests
        num_jobs: Number of parallel threads used for processing (default: 1)

    Returns:
        Dict with keys for each split containing CutSets
    """
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = {}

    # Find all JSONL metadata files
    metadata_files = glob.glob(f"{librispeechmix_metadata_path}/*.jsonl")

    # Collect all dataset parts
    dataset_parts = [Path(f).stem for f in metadata_files]

    # Maybe the manifests already exist
    if output_dir is not None:
        manifests = read_manifests_if_cached(
            dataset_parts=dataset_parts,
            output_dir=output_dir,
            prefix="librispeechmix",
            types=("cutset",),
        )

    for metadata_file in metadata_files:
        part_name = Path(metadata_file).stem

        # Check if manifests already exist and are cached
        if manifests_exist(
            part=part_name,
            output_dir=output_dir,
            prefix="librispeechmix",
            types=("cutset",),
        ):
            logging.info(
                f"LibriSpeechMix subset: {part_name} already prepared - skipping."
            )
            continue

        logging.info(f"Processing {part_name}...")

        librispeech_cutset_path = (
            Path(librispeech_root_path)
            / f"librispeech_cutset_{'-'.join(part_name.split('-')[:-1])}.jsonl.gz"
        )

        librispeech_cutset = CutSet.from_file(librispeech_cutset_path)

        def use_recording_id(cut):
            cut.id = cut.recording_id
            return cut

        librispeech_cutset = librispeech_cutset.map(use_recording_id).to_eager()
        # Process metadata file
        cuts = _process_librispeechmix_metadata(
            metadata_file, librispeech_cutset, num_jobs
        )

        cutset = CutSet.from_cuts(cuts)

        if output_dir is not None:
            cutset.to_file(output_dir / f"librispeechmix_cutset_{part_name}.jsonl.gz")

        manifests[part_name] = {
            "cutset": cutset,
        }

    return manifests


def _process_librispeechmix_metadata(
    metadata_file: str, librispeech_cutset: CutSet, num_jobs: int
) -> List[MonoCut]:
    """Process a LibriSpeechMix metadata JSONL file."""

    # Read JSONL file
    metadata_entries = []
    with open(metadata_file, "r") as f:
        for line in f:
            metadata_entries.append(json.loads(line.strip()))

    cuts = []

    logging.info(f"Processing {len(metadata_entries)} entries from {metadata_file}...")

    with ThreadPoolExecutor(max_workers=num_jobs) as ex:
        futures = [
            ex.submit(_process_metadata_entry, entry, librispeech_cutset)
            for entry in metadata_entries
        ]
        for f in tqdm.tqdm(as_completed(futures), total=len(futures)):
            result = f.result()
            if result is not None:
                cuts.append(result)

    return cuts


def _process_metadata_entry(
    entry: dict, librispeech_cutset: CutSet
) -> Optional[MonoCut]:
    """Process a single metadata entry from LibriSpeechMix."""

    # Extract information from metadata
    mixture_id = entry["id"].split("/")[-1]
    wavs = entry["wavs"]
    delays = entry["delays"]

    # Find corresponding cuts in LibriSpeech
    source_cuts = []
    for i, wav_path in enumerate(wavs):
        # Convert wav path to cut ID (remove extension and convert path separators)
        cut_id = Path(wav_path).stem

        cut = librispeech_cutset[cut_id]

        # Apply delay if specified
        delay = delays[i] if i < len(delays) else 0.0
        if delay > 0:
            cut = cut.pad(delay + cut.duration, direction="left")

        source_cuts.append(cut)

    if len(source_cuts) != len(wavs):
        raise ValueError("Not all mono cuts collected")

    # Create mixture by mixing all source cuts
    mixed_cut = source_cuts[0]
    for source_cut in source_cuts[1:]:
        mixed_cut = mix(mixed_cut, source_cut, preserve_id="left")

    # Update mixture ID
    mixed_cut.id = mixture_id

    return mixed_cut



import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.notsofar1 import download_notsofar1, prepare_notsofar1
from lhotse.utils import Pathlike

__all__ = ["notsofar1"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
def notsofar1(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
):
    """NOTSOFAR-1 data preparation."""
    prepare_notsofar1(
        corpus_dir,
        output_dir=output_dir,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
@click.option(
    "--force-download",
    type=bool,
    is_flag=True,
    default=False,
    help="Force download.",
)
@click.option(
    "-p",
    "--dataset-parts",
    type=str,
    is_flag=False,
    multiple=True,
    default=("train", "dev", "test"),
    help="Dataset parts to download.",
)
@click.option(
    "--mic",
    type=click.Choice(
        ["sdm", "mdm"],
        case_sensitive=False,
    ),
    default="sdm",
    help="NOTSOFAR-1 microphone setting.",
)
@click.option(
    "--train-version",
    type=str,
    is_flag=False,
    default="240825.1_train",
    help="Train dataset version.",
)
@click.option(
    "--dev-version",
    type=str,
    is_flag=False,
    default="240825.1_dev1",
    help="Dev dataset version.",
)
@click.option(
    "--test-version",
    type=str,
    is_flag=False,
    default="240629.1_eval_small_with_GT",
    help="Test dataset version.",
)
def notsofar1(
    target_dir: Pathlike,
    force_download: bool,
    dataset_parts: str,
    mic: str,
    train_version: str,
    dev_version: str,
    test_version: str,
):
    """NOTSOFAR-1 download."""
    download_notsofar1(
        target_dir,
        parts=dataset_parts,
        mic=mic,
        train_version=train_version,
        dev_version=dev_version,
        test_version=test_version,
        force_download=force_download,
    )
